from importlib import import_module

from orchestra.core.errors import MachineExecutionError
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import Step
from orchestra.utils.task_lifecycle import previously_completed_task_data
from orchestra.utils.task_lifecycle import create_subsequent_tasks


def execute(project_id, step_slug):
    project = Project.objects.get(id=project_id)
    step = Step.objects.get(slug=step_slug,
                            workflow_version=project.workflow_version)
    task = Task.objects.get(project=project,
                            step=step)

    # Run machine function
    if step.is_human:
        raise MachineExecutionError('Step worker type is not machine')

    if task.status == Task.Status.COMPLETE:
        raise MachineExecutionError('Task assignment already completed')

    # Machine tasks are only assigned to one worker/machine,
    # so they should only have one task assignment,
    # and should never be submitted for review.
    count = task.assignments.count()
    if count > 1:
        raise MachineExecutionError('At most 1 assignment per machine task')
    elif count == 1:
        task_assignment = task.assignments.first()
        if task_assignment.status == TaskAssignment.Status.SUBMITTED:
            raise MachineExecutionError('Task assignment completed '
                                        'but task is not!')
    else:
        task_assignment = (
            TaskAssignment.objects
            .create(task=task,
                    status=TaskAssignment.Status.PROCESSING,
                    in_progress_task_data={},
                    snapshots={}))

    prerequisites = previously_completed_task_data(task)

    function_module = import_module(step.execution_function['module'])
    function = getattr(function_module, step.execution_function['name'])
    task_data = function(project.project_data, prerequisites)
    task_assignment.status = TaskAssignment.Status.SUBMITTED
    task_assignment.in_progress_task_data = task_data
    task_assignment.save()
    task.status = Task.Status.COMPLETE
    task.save()

    create_subsequent_tasks(project)



import itertools
import numpy as np
from datetime import timedelta
from dateutil.parser import parse
from operator import attrgetter

from pandas import DataFrame

from orchestra.models import Task
from orchestra.project_api.api import get_project_information


def work_time_df(projects, human_only=True, complete_tasks_only=True):
    """
    Return projects' task assignment iteration timing information.

    The returned dataframe's schema is:
        project_id: the project's ID
        project_description: the project's human readable description
        task_id: the task's ID
        task_step_slug: the task's step slug
        assignment_level: a 1-indexed assignment counter (1 is entry-level, 2
            is first review, etc.)
        worker: the username of the worker on that assignment
        iteration: a work iteration for the assignment:
            - Iteration 0 represents the period from previous submission to
            reviewer assignment:
                - If this is the first assignment, 'previous submission'
                  means 'task creation'.
                - For iteration 0, work time will always be 0
                - If this task is auto-assigned, iteration 0's calendar time
                  will be close to 0.
            - Iterations N > 0 represent the period between a previous worker's
                input (submit/accept/reject) and the worker's submission.
        start_datetime: the time the iteration started.
        end_datetime: the time the iteration ended.
        calendar_time: the wallclock time it took for the worker to complete
            that iteration.
        work_time: the amount of work time the worker reported dedicating to
            that iteration (0 for iteration 0).

    Args:
        projects (django.db.models.query.QuerySet):
            An iterable of projects to include in the report.

    Returns:
        df (pandas.DataFrame):
            A DataFrame with timing information.
    """
    # TODO(marcua): `get_project_information` per project means we do random
    # access all over the database.  Ideally, we'd have a sequential scan
    # version of this.
    time_data = list(itertools.chain.from_iterable(
        project_time_row_generator(
            get_project_information(project.id),
            human_only, complete_tasks_only)
        for project in projects))
    df = DataFrame(time_data)
    # Pandas treats all non-primitives as strings, so we explicitly cast
    # datetimes as datetimes instead of strings.
    if not df.empty:
        df['start_datetime'] = df['start_datetime'].astype('datetime64[ns]')
        df['end_datetime'] = df['end_datetime'].astype('datetime64[ns]')
    return df


class Iteration(object):
    """ Represents a task assignment iteration's data for reporting. """
    def __init__(self, project, task, assignment, iteration, worker,
                 work_time, end_datetime):
        self.project = project
        self.task = task
        self.assignment = assignment
        self.iteration = iteration
        self.work_time = work_time
        self.end_datetime = end_datetime
        self.worker = worker

    def to_dict(self, start_datetime):
        return {
            'project_id': self.project['id'],
            'project_description': self.project['short_description'],
            'task_id': self.task['id'],
            'task_step_slug': self.task['step_slug'],
            'assignment_level': self.assignment,
            'iteration': self.iteration,
            'start_datetime': start_datetime,
            'work_time': self.work_time,
            'calendar_time': self.end_datetime - start_datetime,
            'end_datetime': self.end_datetime,
            'worker': self.worker
        }


def project_time_row_generator(project_information,
                               human_only,
                               complete_tasks_only):
    """
    Given a project, yields all task assignment iterations, reporting
    calendar and work times.
    """
    # lol at this variable name
    human_slugs = {
        step['slug'] for step in project_information['steps']
        if step['is_human']}

    for step_slug, task in iter(project_information['tasks'].items()):
        # TODO(marcua): Eventually we'll want to support more fine-grained
        # task filtering for completed vs. in-progress tasks, or to
        # include machine tasks.
        if human_only and step_slug not in human_slugs:
            continue
        if (complete_tasks_only
                and task['status'] !=
                dict(Task.STATUS_CHOICES)[Task.Status.COMPLETE]):
            continue
        iterations = []

        for assignment_idx, assignment in enumerate(task['assignments']):
            iterations.append(Iteration(
                project_information['project'], task,
                assignment_idx + 1, 0, assignment['worker']['username'],
                timedelta(seconds=0),
                parse(assignment['start_datetime'])))
            for iteration, snapshot in enumerate(
                    assignment['snapshots']['snapshots']):
                iterations.append(Iteration(
                    project_information['project'], task,
                    assignment_idx + 1,
                    iteration + 1,
                    assignment['worker']['username'],
                    timedelta(seconds=snapshot['work_time_seconds']),
                    parse(snapshot['datetime'])))
        # To calculate calendar time, zip together each iteration
        # (in time-order) with the end time of the previous iteration
        # so that we can subtract the two.
        iterations = sorted(iterations, key=attrgetter('end_datetime'))
        iteration_times = (
            [parse(task['start_datetime'])] +
            [iteration.end_datetime for iteration in iterations])
        for start_datetime, iteration in zip(iteration_times[:-1], iterations):
            yield iteration.to_dict(start_datetime)


def work_time_sum(df, dimensions):
    """
    Sums work time grouped by `dimensions`.

    Args:
        df (pandas.DataFrame):
            A DataFrame with a schema described in `work_time_df`.
        dimensions (list):
            A list of column names to group on. Example include any subset
            of `['project_id', 'project_description', 'task_id',
            'task_step_slug', 'assignment_level', 'iteration']`.

    Returns:
        df (pandas.DataFrame):
            A DataFrame with timing information.
    """
    return df.groupby(dimensions, as_index=False).agg({'work_time': np.sum})


def calendar_time_sum(df, dimensions):
    """
    Sums calendar time grouped by `dimensions`.

    Args:
        df (pandas.DataFrame):
            A DataFrame with a schema described in `work_time_df`.
        dimensions (list):
            A list of column names to group on. Example include any subset
            of `['project_id', 'project_description', 'task_id',
            'task_step_slug', 'assignment_level', 'iteration']`.

    Returns:
        df (pandas.DataFrame):
            A DataFrame with timing information.
    """
    if (dimensions == ['project_id'] or
            dimensions == ['project_description'] or
            set(dimensions) == {'project_id', 'project_description'}):
        # When reporting the project time, we want to avoid double-counting two
        # tasks that are running in parallel, so we calculate the difference
        # between the farthest-apart start and end times for iterations.
        aggregated = df.groupby(dimensions, as_index=False).agg({
            'start_datetime': np.min,
            'end_datetime': np.max})
        aggregated['calendar_time'] = (
            aggregated['end_datetime'] - aggregated['start_datetime'])
        aggregated.drop('start_datetime', axis=1, inplace=True)
        aggregated.drop('end_datetime', axis=1, inplace=True)
        return aggregated
    else:
        # For groupings that are more granular, like assignments, the work of
        # several workers is interspersed (e.g., worker 1 does work, then
        # worker 2 reviews, etc.). Computing the end time minus the start time
        # would allocate the same calendar time to multiple assignments, so we
        # want to sum up their computed calendar_datetimes.
        return df.groupby(dimensions, as_index=False).agg(
            {'calendar_time': np.sum})



from django.contrib.auth.models import User
from django.db import models
from django.utils import timezone
from jsonfield import JSONField

from orchestra.core.errors import ModelSaveError
from orchestra.utils.assignment_snapshots import load_snapshots

# TODO(marcua): Convert ManyToManyFields to django-hstore referencefields or
# wait for django-postgres ArrayFields in Django 1.8.


class Workflow(models.Model):
    """
    Workflows describe the steps and requirements for experts to complete work.

    This model represents workflows that have been loaded into the database and
    are usable by Orchestra projects. All workflows are also defined in the
    codebase (see the `code_directory` attribute).

    Attributes:
        slug (str):
            Unique identifier for the workflow.
        name (str):
            Human-readable name for the workflow.
        description (str):
            A longer description of the workflow.
        code_directory (str):
            The full path to the location of the workflow's manifest.
    """
    slug = models.CharField(max_length=200, unique=True)
    name = models.CharField(max_length=200)
    description = models.TextField()
    code_directory = models.CharField(max_length=255, unique=True)
    sample_data_load_function = JSONField(default={})

    class Meta:
        app_label = 'orchestra'

    def __str__(self):
        return self.slug


class WorkflowVersion(models.Model):
    """
    WorkflowVersions represent changes made to a single workflow over time.

    Attributes:
        slug (str):
            Unique identifier for the workflow version.
        name (str):
            Human-readable name for the workflow version.
        description (str):
            A longer description of the workflow version.
        workflow (orchestra.models.Workflow):
            The workflow that this is a version of.
    """
    slug = models.CharField(max_length=200)
    name = models.CharField(max_length=200)
    description = models.TextField()
    workflow = models.ForeignKey(Workflow, related_name='versions')

    class Meta:
        app_label = 'orchestra'
        unique_together = (('workflow', 'slug'),)

    def __str__(self):
        return '{} - {}'.format(self.workflow.slug, self.slug)


class Certification(models.Model):
    """
    Certifications allow workers to perform different types of tasks.

    Attributes:
        slug (str):
            Unique identifier for the certification.
        name (str):
            Human-readable name for the certification.
        description (str):
            A longer description of the certification.
        required_certifications ([orchestra.models.Certification]):
            Prerequisite certifications for possessing this one.
        workflow (orchestra.models.Workflow):
            The workflow the certification is scoped to.
    """
    slug = models.CharField(max_length=200)
    name = models.CharField(max_length=200)
    description = models.TextField()
    required_certifications = models.ManyToManyField(
        'self',
        blank=True,
        related_name='dependent_certifications',
        symmetrical=False)
    workflow = models.ForeignKey(Workflow, related_name='certifications')

    def __str__(self):
        return '{} - {}'.format(self.slug, self.workflow.slug)

    class Meta:
        app_label = 'orchestra'
        unique_together = (('workflow', 'slug'),)


class Step(models.Model):
    """
    Steps represent individual tasks in a workflow version.

    Attributes:
        slug (str):
            Unique identifier for the workflow step.
        name (str):
            Human-readable name for the workflow step.
        description (str):
            A longer description of the workflow step.
        workflow_version (orchestra.models.WorkflowVersion):
            The workflow version that this is a step of.
        creation_depends_on ([orchestra.models.Step]):
            Workflow steps that must be complete to begin this step.
        submission_depends_on ([orchestra.models.Step]):
            Workflow steps that must be complete to end this step.
        is_human (bool):
            False if this step is performed by a machine.
        execution_function (str):
            A JSON blob containing the path to and name of a python method that
            will execute this step (only valid for machine steps).
        required_certifications ([orchestra.models.Certification]):
            The certifications a worker must have in order to work on this step
            (only valid for human steps).
        assignment_policy (str):
            A JSON blob used to decide which worker to assign to this step
            (only valid for human steps).
        review_policy (str):
            A JSON blob used to decide whether or not to review this step
            (only valid for human steps).
        user_interface (str):
            A JSON blob used to describe the files used in the user interface
            for this step (only valid for human steps).
    """
    # General fields
    slug = models.CharField(max_length=200)
    name = models.CharField(max_length=200)
    description = models.TextField()
    workflow_version = models.ForeignKey(WorkflowVersion, related_name='steps')
    creation_depends_on = models.ManyToManyField(
        'self',
        blank=True,
        related_name='creation_dependents',
        symmetrical=False)
    submission_depends_on = models.ManyToManyField(
        'self',
        blank=True,
        related_name='submission_dependents',
        symmetrical=False)

    # Machine step fields
    is_human = models.BooleanField()
    execution_function = JSONField(default={})

    # Human step fields
    required_certifications = models.ManyToManyField(Certification, blank=True)
    assignment_policy = JSONField(default={})
    review_policy = JSONField(default={})
    user_interface = JSONField(default={})

    class Meta:
        app_label = 'orchestra'
        unique_together = (('workflow_version', 'slug'),)

    def __str__(self):
        return '{} - {} - {}'.format(self.slug, self.workflow_version.slug,
                                     self.workflow_version.workflow.slug)


class Worker(models.Model):
    """
    Workers are human experts within the Orchestra ecosystem.

    Attributes:
        user (django.contrib.auth.models.User):
            Django user whom the worker represents.
        start_datetime (datetime.datetime):
            The time the worker was created.
        slack_username (str):
            The worker's Slack username if Slack integration is enabled.
    """
    user = models.OneToOneField(User)
    start_datetime = models.DateTimeField(default=timezone.now)
    slack_username = models.CharField(max_length=200, blank=True, null=True)

    def __str__(self):
        return '{}'.format(self.user.username)

    class Meta:
        app_label = 'orchestra'


class WorkerCertification(models.Model):
    """
    A WorkerCertification maps a worker to a certification they possess.

    Attributes:
        certification (orchestra.models.Certification):
            Certification belonging to the corresponding worker.
        worker (orchestra.models.Worker):
            Worker possessing the given certification.
        task_class (orchestra.models.WorkerCertification.TaskClass):
            Represents whether the worker is in training for the given
            certification or prepared to work on real tasks.
        role (orchestra.models.WorkerCertification.Role):
            Represents whather the worker is an entry-level or review
            worker for the given certification.

    Constraints:
        `certification`, `worker`, `task_class`, and `role` are taken
        to be unique_together.

        Worker must possess an entry-level WorkerCertification before
        obtaining a reviewer one.
    """
    class Meta:
        app_label = 'orchestra'
        unique_together = ('certification', 'worker', 'task_class', 'role')

    class TaskClass:
        TRAINING = 0
        REAL = 1

    TASK_CLASS_CHOICES = (
        (TaskClass.TRAINING, 'Training tasks'),
        (TaskClass.REAL, 'A real task'))

    # If a worker has a REVIEWER certification, then they must have
    # an ENTRY_LEVEL certification
    class Role:
        ENTRY_LEVEL = 0
        REVIEWER = 1

    ROLE_CHOICES = (
        (Role.ENTRY_LEVEL, 'Entry-level'),
        (Role.REVIEWER, 'Reviewer'))

    certification = models.ForeignKey(Certification)
    worker = models.ForeignKey(Worker, related_name='certifications')
    task_class = models.IntegerField(choices=TASK_CLASS_CHOICES)
    role = models.IntegerField(choices=ROLE_CHOICES)

    def __str__(self):
        return '{} - {} - {} - {} - {}'.format(
            self.worker.user.username, self.certification.slug,
            self.certification.workflow.slug,
            dict(WorkerCertification.TASK_CLASS_CHOICES)[self.task_class],
            dict(WorkerCertification.ROLE_CHOICES)[self.role])

    def save(self, *args, **kwargs):
        if self.role == WorkerCertification.Role.REVIEWER:
            if not (WorkerCertification.objects
                    .filter(worker=self.worker, task_class=self.task_class,
                            certification=self.certification,
                            role=WorkerCertification.Role.ENTRY_LEVEL)
                    .exists()):
                raise ModelSaveError('You are trying to add a reviewer '
                                     'certification ({}) for a worker without '
                                     'an entry-level certification'
                                     .format(self))
        super(WorkerCertification, self).save(*args, **kwargs)


class Project(models.Model):
    """
    A project is a collection of tasks representing a workflow.

    Attributes:
        start_datetime (datetime.datetime):
            The time the project was created.
        status (orchestra.models.Project.Status):
            Represents whether the project is being actively worked on.
        workflow_version (orchestra.models.WorkflowVersion):
            Identifies the workflow that the project follows.
        priority (int):
            Represents the relative priority of the project.
        task_class (int):
            Represents whether the project is a worker training exercise
            or a deliverable project.
        review_document_url (str):
            The URL for the review document to be passed between workers
            and reviwers for the project's tasks.
        slack_group_id (str):
            The project's internal Slack group ID if Slack integration
            is enabled.
    """
    class Status:
        ACTIVE = 0
        ABORTED = 2

    STATUS_CHOICES = (
        (Status.ACTIVE, 'Active'),
        (Status.ABORTED, 'Aborted'))

    start_datetime = models.DateTimeField(default=timezone.now)
    status = models.IntegerField(choices=STATUS_CHOICES,
                                 default=Status.ACTIVE)

    workflow_version = models.ForeignKey(WorkflowVersion,
                                         related_name='projects')

    short_description = models.TextField()
    priority = models.IntegerField()
    project_data = JSONField(default={}, blank=True)
    task_class = models.IntegerField(
        choices=WorkerCertification.TASK_CLASS_CHOICES)
    review_document_url = models.URLField(null=True, blank=True)
    slack_group_id = models.CharField(max_length=200, null=True, blank=True)

    def __str__(self):
        return '{} ({})'.format(str(self.workflow_version.slug),
                                self.short_description)

    class Meta:
        app_label = 'orchestra'


class Task(models.Model):
    """
    A task is a cohesive unit of work representing a workflow step.

    Attributes:
        start_datetime (datetime.datetime):
            The time the Task was created.
        step (orchestra.models.Step):
            Identifies the step that the task represents.
        project (orchestra.models.Project):
            The project to which the task belongs.
        status (orchestra.models.Task.Status):
            Represents the task's stage within its lifecycle.
    """
    class Status:
        AWAITING_PROCESSING = 0
        PROCESSING = 1
        PENDING_REVIEW = 2
        REVIEWING = 3
        POST_REVIEW_PROCESSING = 4
        COMPLETE = 5
        ABORTED = 6

    STATUS_CHOICES = (
        (Status.AWAITING_PROCESSING, 'Awaiting Processing'),
        (Status.PROCESSING, 'Processing'),
        (Status.PENDING_REVIEW, 'Pending Review'),
        (Status.REVIEWING, 'Reviewing'),
        (Status.POST_REVIEW_PROCESSING, 'Post-review Processing'),
        (Status.ABORTED, 'Aborted'),
        (Status.COMPLETE, 'Complete'))

    start_datetime = models.DateTimeField(default=timezone.now)
    step = models.ForeignKey(Step, related_name='tasks')
    project = models.ForeignKey(Project, related_name='tasks')
    status = models.IntegerField(choices=STATUS_CHOICES)

    def __str__(self):
        return '{} - {}'.format(str(self.project), str(self.step.slug))

    class Meta:
        app_label = 'orchestra'


class TaskAssignment(models.Model):
    """
    A task assignment is a worker's assignment for a given task.

    Attributes:
        start_datetime (datetime.datetime):
            The time the task was assigned for this iteration.
        worker (orchestra.models.Worker):
            The worker to whom the given task is assigned.
        task (orchestra.models.Task):
            The given task for the task assignment.
        status (orchestra.models.Project.Status):
            Represents whether the assignment is currently being worked
            on.
        assignment_counter (int):
            Identifies the level of the assignment in the given task's
            review hierarchy (i.e., 0 represents an entry-level worker,
            1 represents the task's first reviewer, etc.).
        in_progress_task_data (str):
            A JSON blob containing the worker's input data for the task
            assignment.
        snapshots (str):
            A JSON blob containing saved snapshots of previous data from
            the task assignment.

    Constraints:
        `task` and `assignment_counter` are taken to be unique_together.

        Task assignments for machine-type tasks cannot have a `worker`,
        while those for human-type tasks must have one.
    """
    class Meta:
        app_label = 'orchestra'
        unique_together = ('task', 'assignment_counter')

    class SnapshotType:
        SUBMIT = 0
        ACCEPT = 1
        REJECT = 2

    class Status:
        PROCESSING = 0
        SUBMITTED = 1

    STATUS_CHOICES = (
        (Status.PROCESSING, 'Processing'),
        (Status.SUBMITTED, 'Submitted'))

    start_datetime = models.DateTimeField(default=timezone.now)
    worker = models.ForeignKey(Worker,
                               null=True,
                               blank=True)
    task = models.ForeignKey(Task, related_name='assignments')

    status = models.IntegerField(choices=STATUS_CHOICES)

    # Counter of a worker assigned to the task
    assignment_counter = models.IntegerField(default=0)

    # Opaque field that stores current state of task as per the Step's
    # description
    in_progress_task_data = JSONField(default={}, blank=True)

    # When a worker submits, accepts, or rejects a task, we snapshot their
    # in_workflow_task_data along with the date in the following format:
    # {'snapshots': [
    #   {'data': snapshotted_task_data,
    #    'datetime': ISO 8601 datetime in UTC time,
    #    'work_time_seconds': integer seconds,
    #    'type': value from SnapshotType}]
    #  '__version': 1}
    snapshots = JSONField(default={}, blank=True)

    def save(self, *args, **kwargs):
        if self.task.step.is_human:
            if self.worker is None:
                raise ModelSaveError('Worker has to be present '
                                     'if worker type is Human')
        else:
            if self.worker is not None:
                raise ModelSaveError('Worker should not be assigned '
                                     'if worker type is Machine')

        super(TaskAssignment, self).save(*args, **kwargs)

    def __str__(self):
        return '{} - {} - {}'.format(
            str(self.task), self.assignment_counter, str(self.worker))


# Attach a post-init signal to TaskAssigment.  Every
# TaskAssignment that gets constructed will now call
# this post-init signal after loading from the database
# (or memory).  We run `load_snapshots` after loading from
# the database so that we can migrate old JSON task assignment
# snapshots.
def task_assignment_post_init(sender, instance, **kwargs):
    instance.snapshots = load_snapshots(instance.snapshots)
models.signals.post_init.connect(
    task_assignment_post_init, sender=TaskAssignment)



from django.contrib import admin

from orchestra.models import Certification
from orchestra.models import Worker
from orchestra.models import WorkerCertification
from orchestra.models import Project
from orchestra.models import Step
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import Workflow
from orchestra.models import WorkflowVersion

admin.site.register(Certification)
admin.site.register(Worker)
admin.site.register(WorkerCertification)
admin.site.register(Project)
admin.site.register(Step)
admin.site.register(Task)
admin.site.register(TaskAssignment)
admin.site.register(Workflow)
admin.site.register(WorkflowVersion)

admin.site.site_header = 'Orchestra'
admin.site.site_title = 'Orchestra'
admin.site.index_title = 'Orchestra'



from copy import deepcopy
from dateutil.parser import parse
from dateutil.tz import tzutc

from orchestra.core.errors import SnapshotsError


VERSION = '__version'
SNAPSHOTS = 'snapshots'
DATETIME = 'datetime'
TIME_WORKED = 'work_time_seconds'


def empty_snapshots():
    return {SNAPSHOTS: [],
            VERSION: 2}


def load_snapshots(snapshots):
    """
    Given a task assignment `snapshots` object, ensures that it follows
    the latest snapshots schema.
    """
    snapshots = deepcopy(snapshots)
    version = snapshots.get(VERSION)
    if version is None:
        snapshot_list = snapshots.get(SNAPSHOTS, [])
        for snapshot_data in snapshot_list:
            snapshot_data[TIME_WORKED] = snapshot_data.get(TIME_WORKED, 0)
        snapshots[SNAPSHOTS] = snapshot_list
        snapshots[VERSION] = 1
    if snapshots[VERSION] == 1:
        snapshot_list = snapshots.get(SNAPSHOTS, [])
        for snapshot_data in snapshot_list:
            snapshot_data[DATETIME] = parse(
                snapshot_data[DATETIME]).replace(tzinfo=tzutc()).isoformat()
        snapshots[SNAPSHOTS] = snapshot_list
        snapshots[VERSION] = 2
    if snapshots[VERSION] != 2:
        raise SnapshotsError('Task Assignment snapshots were an '
                             'unrecognized version: {}'.format(snapshots))
    return snapshots



from django.conf.urls import include
from django.conf.urls import patterns
from django.conf.urls import url
from orchestra.views import index
from orchestra.views import status


urlpatterns = patterns(
    '',
    url(r'^api/',
        include('orchestra.api_urls', namespace='orchestra')),
    url(r'^app/?', index, name='index'),
    url(r'^status/', status, name='status')
)



from urllib.parse import urljoin

from django.conf import settings
from django.core import urlresolvers
from django.utils import timezone

from orchestra.analytics.latency import work_time_df
from orchestra.models import Project
from orchestra.project_api.api import get_project_information
from orchestra.slack import SlackService
from orchestra.utils.assignment_snapshots import empty_snapshots
from orchestra.utils.task_properties import last_snapshotted_assignment

import logging

logger = logging.getLogger(__name__)


def project_management_information(project_id):
    project = Project.objects.get(id=project_id)
    df = work_time_df([project],
                      human_only=False, complete_tasks_only=False)
    df['end_datetime'] = df['start_datetime'] + df['calendar_time']
    project_information = get_project_information(project.id)
    project_information['project']['status'] = dict(
        Project.STATUS_CHOICES).get(project.status, None)
    project_information['project']['admin_url'] = urljoin(
        settings.ORCHESTRA_URL,
        urlresolvers.reverse(
            'admin:orchestra_project_change',
            args=(project_id,)))

    for slug, task in project_information['tasks'].items():
        task['admin_url'] = urljoin(
            settings.ORCHESTRA_URL,
            urlresolvers.reverse(
                'admin:orchestra_task_change',
                args=(task['id'],)))

        for assignment in task['assignments']:
            assignment['admin_url'] = urljoin(
                settings.ORCHESTRA_URL,
                urlresolvers.reverse(
                    'admin:orchestra_taskassignment_change',
                    args=(assignment['id'],)))

            iterations = df[(df.worker == assignment['worker']['username']) &
                            (df.task_id == task['id'])]
            iterations = iterations[['start_datetime', 'end_datetime']]
            assignment['iterations'] = []
            for idx, info in iterations.T.items():
                iteration = info.to_dict()
                assignment['iterations'].append(iteration)
            if assignment['status'] == 'Processing':
                last_iteration_end = assignment['start_datetime']
                last_assignment = last_snapshotted_assignment(task['id'])
                if last_assignment and len(assignment['iterations']) > 1:
                    last_iteration_end = (
                        last_assignment.snapshots['snapshots'][-1]['datetime'])

                assignment['iterations'].append({
                    'start_datetime': last_iteration_end,
                    'end_datetime': timezone.now()
                })

        if task['status'] in ('Awaiting Processing', 'Pending Review'):
            last_assignment_end = task['start_datetime']
            last_assignment = last_snapshotted_assignment(task['id'])
            if last_assignment:
                last_assignment_end = (
                    last_assignment.snapshots['snapshots'][-1]['datetime'])
            task['assignments'].append({
                'iterations': [{
                    'start_datetime': last_assignment_end,
                    'end_datetime': timezone.now()
                }],
                'snapshots': empty_snapshots(),
                'start_datetime': last_assignment_end,
                'status': 'Processing',
                'task': task['id'],
                'worker': {'id': None, 'username': None},
            })
    return project_information


def edit_slack_membership(project_id, username, action):
    slack = SlackService(settings.SLACK_EXPERTS_API_KEY)
    slack_user_id = slack.users.get_user_id(username)
    slack_group_id = Project.objects.get(id=project_id).slack_group_id
    if action == 'add':
        slack.groups.invite(slack_group_id, slack_user_id)
    elif action == 'remove':
        slack.groups.kick(slack_group_id, slack_user_id)
    else:
        raise Exception('Action not found.')



from django.dispatch import receiver
from registration.signals import user_registered

from orchestra.models import Worker


@receiver(user_registered)
def add_worker_for_new_users(sender, user, request, **kwargs):
    Worker.objects.create(user=user)



from django.conf.urls import patterns
from django.conf.urls import url
from beanstalk_dispatch.views import dispatcher

urlpatterns = patterns(
    '',
    url(r'^', dispatcher, name='beanstalk_dispatcher'))



"""
Display a project latency report.

TODO(marcua): Fill the script in.  It's currently a dummy placeholder
with dummy placeholder arguments.
"""

from django.core.management.base import BaseCommand
from orchestra.analytics.latency import work_time_df
from orchestra.models import Project


import logging
logger = logging.getLogger(__name__)


def main(options):
    print(work_time_df(Project.objects.filter(id=1)))


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('--remove-fake',
                            action='store_true',
                            dest='remove_fake',
                            default=False,
                            help='Remove all fake projects and tasks')

    def handle(self, *args, **options):
        main(options)



from registration.signals import user_registered

from orchestra.core.errors import ModelSaveError
from orchestra.models import Worker
from orchestra.models import WorkerCertification
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import CertificationFactory
from orchestra.tests.helpers.fixtures import WorkerCertificationFactory
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.tests.helpers.fixtures import setup_models


class ModelsTestCase(OrchestraTestCase):

    def setUp(self):
        super(ModelsTestCase, self).setUp()
        setup_models(self)
        self.workers[0] = Worker.objects.get(user__username='test_user_1')

    def test_certification_roles(self):
        """ Ensure that workers can be certified at multiple roles. """
        certification = CertificationFactory(
            slug='cat_herding', name='Cat herding',
            workflow=self.workflows['w1'])
        WorkerCertificationFactory(
            certification=certification,
            worker=self.workers[0],
            role=WorkerCertification.Role.ENTRY_LEVEL)
        WorkerCertificationFactory(
            certification=certification,
            worker=self.workers[0],
            role=WorkerCertification.Role.REVIEWER)

    def test_reviewer_requires_entry_level(self):
        """ Workers must have entry-level certification before they review. """
        certification = CertificationFactory(
            slug='cat_herding', name='Cat herding',
            workflow=self.workflows['w1'])
        with self.assertRaises(ModelSaveError):
            WorkerCertificationFactory(
                certification=certification,
                worker=self.workers[0],
                role=WorkerCertification.Role.REVIEWER)

    def test_worker_autocreation(self):
        """ When new users register, worker objects should be autocreated. """

        # Create a user
        user = UserFactory(username='test_registration_user',
                           password='test',
                           email='test_registration_user@test.com')

        # There should be no worker yet
        self.assertFalse(Worker.objects.filter(user=user).exists())

        # Fake registering the user
        user_registered.send(sender=self.__class__, user=user, request=None)

        # Expect the worker object to be created
        self.assertTrue(Worker.objects.filter(user=user).exists(),
                        'Worker not autocreated on User registration')
